/*
 * Copyright (C) 2009 Alan Hightower
 * Sigma Designs smp86xx SoC mixer driver
 * http://www.musicpd.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <unistd.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <sys/times.h>

#include <glib.h>

#include "conf.h"
#include "mixer_api.h"

// Sigma includes
#define ALLOW_OS_CODE 1
#define EM86XX_CHIP EM86XX_CHIPID_TANGO2
#define EM86XX_MODE EM86XX_MODEID_STANDALONE
#define EM86XX_REVISION 6

#include <rmdef/rmdef.h>
#include <rmdef/rmdbg.h>
#include <rmcore/include/rmstatustostring.h>

#include <rua/include/rua.h>
#include <rua/include/rua_property.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "smp86xx_mixer"

#define DEFAULT_ENGINE      0
#define AUDIO_ENGINE_MAX_CHANNELS  12 // 7.1 + stereo downmix + s/p-dif

extern const struct mixer_plugin smp86xx_mixer;

typedef struct _st_smp86xx_mixer
{
	/**
	 * Base mixer class
	 */
	struct mixer base;

	/**
	 * Sigma emhwlib rpc handle
	 */
	struct RUA *prua;

	/**
	 * Module ID of the DSP used for audio output
	 */
	RMuint32 engine_id;

	/**
	 * Current normalized volume
	 */
	int current;

} smp86xx_mixer_t;

// Helpers for passing property args by reference
#define PROPERTY_ARG_REF(a) &a, sizeof (a)

/*
 * Work around for a bug in the MPD framework that calls
 * get volume without ever calling open
 */
#define BROKEN_MPD_MIXER_OPEN  1


/** Configure and initialize mixer
 *
 * Initializes mixer context block and selects which DSP should be
 * the target of control
 *
 * @param param the configuration section, or NULL if there is
 * no configuration
 * @return NULL on error, or a pointer to mixer context
 */
static struct mixer *smp86xx_mixer_init
(
	const struct config_param *param
)
{
	smp86xx_mixer_t *mixer = g_new (smp86xx_mixer_t, 1);

// --------------------------------------------------------------------

	memset (mixer, 0, sizeof (smp86xx_mixer_t));

	mixer->current = 100;

	mixer_init (&mixer->base, &smp86xx_mixer);

	mixer->engine_id =
		EMHWLIB_MODULE (AudioEngine,
		                config_get_block_unsigned (param, "engine", DEFAULT_ENGINE));

	return &mixer->base;

} // end of "smp86xx_mixer_init ()"


// mute -> 0db gain normalized linear volume table
static const RMuint32 volume_table[] = {
	0x00000000, // mute
	0x00100000, 0x0011f59c, 0x001428a4, 0x0016a09f, 0x001965ff, 0x001c823e, // -48dB ... -43dB
	0x00200000, 0x0023eb35, 0x00285145, 0x002d413c, 0x0032cbfc, 0x0039047b, // -42dB ... -37dB
	0x00400000, 0x0047d66a, 0x0050a28b, 0x005a8279, 0x006597fa, 0x007208f8, // -36dB ... -31dB
	0x00800000, 0x008facd6, 0x00a14518, 0x00b504f3, 0x00cb2ff5, 0x00e411f0, // -30dB ... -25dB
	0x01000000, 0x011f59ac, 0x01428a2f, 0x016a09e6, 0x01965fea, 0x01c823e0, // -24dB ... -19dB
	0x02000000, 0x023eb358, 0x0285145f, 0x02d413cd, 0x032cbfd5, 0x039047c1, // -18dB ... -13dB
	0x04000000, 0x047d66b1, 0x050a28be, 0x05a82799, 0x06597fa9, 0x07208f82, // -12dB ...  -7dB
	0x08000000, 0x08facd62, 0x0a14517d, 0x0b504f34, 0x0cb2ff53, 0x0e411f04, //  -6dB ...  -1dB
	0x10000000, // 0 db
};

#define table_entries (sizeof (volume_table) / sizeof (volume_table[0]))
#define VOL_MAX  0x10000000


/** Open mixer
 *
 * @param data Mixer context block
 * @return false on error, true on success
 */
static bool smp86xx_mixer_open
(
	struct mixer *data
)
{
	int v;

	RMstatus sts;
	unsigned new_volume;

	struct AudioEngine_Volume_type volume;

	smp86xx_mixer_t *mixer = (smp86xx_mixer_t *) data;

// --------------------------------------------------------------------

	// Open a RPC handle to Sigma API

	sts = RUACreateInstance (&mixer->prua, 0); // board 0

	if (RMFAILED (sts))
	{
		g_error ("Unable to open RUA handle - %s", RMstatusToString (sts));

		return false;
	}

	new_volume = (mixer->current * (table_entries - 1)) / 100;

	for (v = 0; v < AUDIO_ENGINE_MAX_CHANNELS; v++)
	{
		volume.Channel = v;
		volume.Volume  = volume_table[new_volume];

		RUASetProperty (mixer->prua, mixer->engine_id,
		                RMAudioEnginePropertyID_Volume,
		                PROPERTY_ARG_REF (volume), 0);
	}

	return true;

} // end of "smp86xx_mixer_open ()"


/** Return the current volume.
 *
 * @param data Mixer context block
 * @return the current volume (0..100 including) or -1 on
 * error
 */
static int smp86xx_mixer_get_volume
(
	struct mixer *data
)
{
	smp86xx_mixer_t *mixer = (smp86xx_mixer_t *) data;

// --------------------------------------------------------------------

#if BROKEN_MPD_MIXER_OPEN
	if (!mixer->prua)
		smp86xx_mixer_open (data);
#endif
	return mixer->current;

} // end of "smp86xx_mixer_get_volume ()"


/** Sets the volume.
 *
 * @param data Mixer context block
 * @param new_volume the new volume (0..100 including)
 * @return true on success
 */
static bool smp86xx_mixer_set_volume
(
	struct mixer *data,
	unsigned new_volume
)
{
	int v;

	struct AudioEngine_Volume_type volume;

	smp86xx_mixer_t *mixer = (smp86xx_mixer_t *) data;

// --------------------------------------------------------------------

	mixer->current = new_volume;

	new_volume = (new_volume * (table_entries - 1)) / 100;

	for (v = 0; v < AUDIO_ENGINE_MAX_CHANNELS; v++)
	{
		volume.Channel = v;
		volume.Volume  = volume_table[new_volume];

		RUASetProperty (mixer->prua, mixer->engine_id,
		                RMAudioEnginePropertyID_Volume,
		                PROPERTY_ARG_REF (volume), 0);
	}

	return true;

} // end of "smp86xx_mixer_set_volume ()"


/** Close mixer device
 *
 * @param data Mixer context block
 */
static void smp86xx_mixer_close
(
	struct mixer *data
)
{
	smp86xx_mixer_t *mixer = (smp86xx_mixer_t *) data;

// --------------------------------------------------------------------

	if (mixer->prua)
	{
		RUADestroyInstance (mixer->prua);

		mixer->prua = NULL;
	}

	return;

} // end of "smp86xx_mixer_close ()"


/** Terminate access to smp86xx mixer
 *
 * Frees mixer context and associated resources
 *
 * @param data Mixer context block
 */
static void smp86xx_mixer_term
(
	struct mixer *data
)
{
	smp86xx_mixer_t *mixer = (smp86xx_mixer_t *) data;

// --------------------------------------------------------------------

	g_free (mixer);

	return;

} // end of "smp86xx_mixer_term ()"


const struct mixer_plugin smp86xx_mixer =
{
	.init       = smp86xx_mixer_init,
	.open       = smp86xx_mixer_open,
	.close      = smp86xx_mixer_close,
	.finish     = smp86xx_mixer_term,
	.get_volume = smp86xx_mixer_get_volume,
	.set_volume = smp86xx_mixer_set_volume,
	.global     = true,
};
