/*
 * Copyright (C) 2009 Alan Hightower
 * Sigma Designs smp86xx SoC PCM output driver
 * http://www.musicpd.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
 * TODO:
 * - Work-around MPDs broken 24-bit sample packing (or lack of)
 * - libmad compiled for accuracy overdrives 16-bit sampling
 * - HDMI support for internal SiI9030 and external SiI9134
 * - Enable Prologic IIx for players with license
 */

#include <unistd.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <sys/times.h>

#include <glib.h>

#include "conf.h"
#include "output_api.h"

// Sigma includes
#define ALLOW_OS_CODE 1
#define EM86XX_CHIP EM86XX_CHIPID_TANGO2
#define EM86XX_MODE EM86XX_MODEID_STANDALONE
#define EM86XX_REVISION 6

#include <rmdef/rmdef.h>
#include <rmdef/rmdbg.h>
#include <rmcore/include/rmstatustostring.h>

#include <rua/include/rua.h>
#include <rua/include/rua_property.h>

#include <gbuslib/include/gbus_fifo.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "smp86xx_audio"

#define DEFAULT_DRAM_BANK   0
#define DEFAULT_ENGINE      0
#define DEFAULT_DECODER     0

#define DEFAULT_PCM_LINE_COUNT       0xf00
#define DEFAULT_BITSTREAM_FIFO_SIZE  500000 // in us

#define PENDING_RETRY_TIMEOUT   1000 // in ms
#define STC_RESOLUTION         90000

// Only external 9134 on GPIO 0/1 is supported atm
#define I2C_MODULE       0 // 0=ext bitbang, 1=ext hw block, 2=int hw block
#define I2C_BUS_SPEED  100 // in kHz
#define I2C_BIT_DELAY    3 // in usec
#define HDMI_BUS_ADDR 0x72

typedef struct _st_smp86xx_audio
{
	/**
	 * Sigma emhwlib rpc handle
	 */
	struct RUA *prua;

	/**
	 * Set property wait timeout in kernel jiffies
	 */
	clock_t sp_timeout;

	/**
	 * Module ID of the STC timer used for audio sample pacing
	 */
	RMuint32 stc_id;

	/**
	 * Audio timer ID
	 */
	int timer_id;

	/**
	 * Module ID of the DSP used for audio output
	 */
	RMuint32 engine_id;

	/**
	 * Module ID of the logical decoder task within the DSP
	 */
	RMuint32 decoder_id;

	/**
	 * RAMDAC used when allocating audio decoder memory
	 */
	int dram_bank;

	/**
	 * Number of PCM lines propagated to the output sample
	 * buffer at a time
	 */
	unsigned int pcm_line_count;

	/**
	 * Size of the input sample buffer in microseconds
	 */
	uint32_t fifo_duration;

	/**
	 * Size of the input sample buffer in bytes (computed
	 * from duration and audio format)
	 */
	uint32_t fifo_size;

	/**
	 * True if the decoder is currently open
	 */
	bool opened;

	/**
	 * True if the decoder is playing, false if paused or stopped
	 */
	bool playing;

	/**
	 * gbus address of the start of the audio decoder context
	 */
	RMuint32 mem_address;

	/**
	 * Audio decoder context size
	 */
	RMuint32 mem_size;

	/**
	 * Local virtual memory address of audio decoder context
	 */
	uint8_t *mem_ptr;

	/**
	 * DSP shared memory address
	 */
	uint32_t shared_address;

	/**
	 * DSP shared memory size
	 */
	uint32_t shared_size;

	/**
	 * gbus fifo container for audio input buffer
	 */
	volatile struct gbus_fifo *fifo;

	/**
	 * Pointer to base address of input sample fifo
	 */
	int8_t *input;

	/**
	 * HDMI output driver support
	 */
	const char *hdmi_output;

} smp86xx_audio_t;


static inline RMstatus _setdp
(
	smp86xx_audio_t *audio,
	RMuint32 ModuleID,
	RMuint32 PropertyID, 
	void *pValue,
	RMuint32 ValueSize
)
{
	RMstatus sts;

	clock_t timeout = times (NULL) + audio->sp_timeout;

// --------------------------------------------------------------------

	do
	{
		sts = RUASetProperty (audio->prua, ModuleID, PropertyID,
		                      pValue, ValueSize, 0);

	} while ((sts == RM_PENDING) && (times (NULL) < timeout));

	return sts;

} // end of "_setdp ()"


// Helpers for passing property args by reference
#define PROPERTY_ARG_REF(a) &a, sizeof (a)
#define SETDP(m,p,a) _setdp(audio, m, p, &a, sizeof (a))


static inline GQuark _output_quark (void)
{
	return g_quark_from_static_string ("smp86xx_output");
}


static void i2c_write (struct RUA *prua, int offset, int addr, uint8_t data)
{
	struct EMhwlibI2CDeviceParameter params;

	struct I2C_WriteRMuint8_type arg;

// --------------------------------------------------------------------

	params.APIVersion = 1;
	params.Clock   = GPIOId_Sys_0;
	params.Data    = GPIOId_Sys_1;
	params.DevAddr = HDMI_BUS_ADDR + offset;
	params.Delay   = I2C_BIT_DELAY;
	params.Speed   = I2C_BUS_SPEED;

	RUASetProperty (prua, EMHWLIB_MODULE (I2C, I2C_MODULE),
	                RMI2CPropertyID_DeviceParameter,
	                PROPERTY_ARG_REF (params), 0);

	arg.SubAddr = addr;
	arg.Data = data;

	RUASetProperty (prua, EMHWLIB_MODULE (I2C, I2C_MODULE),
	                RMI2CPropertyID_WriteRMuint8,
	                PROPERTY_ARG_REF (arg), 0);
	return;

} // end of "i2c_write ()"


static uint8_t i2c_read (struct RUA *prua, int offset, int addr)
{
	struct EMhwlibI2CDeviceParameter params;

	struct I2C_QueryRMuint8_in_type arg_in;
	struct I2C_QueryRMuint8_out_type arg_out;

// --------------------------------------------------------------------

	params.APIVersion = 1;
	params.Clock   = GPIOId_Sys_0;
	params.Data    = GPIOId_Sys_1;
	params.DevAddr = HDMI_BUS_ADDR + offset;
	params.Delay   = I2C_BIT_DELAY;
	params.Speed   = I2C_BUS_SPEED;

	RUASetProperty (prua, EMHWLIB_MODULE (I2C, I2C_MODULE),
	                RMI2CPropertyID_DeviceParameter,
	                PROPERTY_ARG_REF (params), 0);

	arg_in.SubAddr = addr;
	arg_out.Data   = 0;

	RUAExchangeProperty (prua, EMHWLIB_MODULE (I2C, I2C_MODULE),
	                     RMI2CPropertyID_QueryRMuint8,
	                     PROPERTY_ARG_REF (arg_in),
	                     PROPERTY_ARG_REF (arg_out));

	return arg_out.Data;

} // end of "i2c_read ()"


#define preg(a,b)										  \
    g_debug ("preg %d+0x%02x=0x%02x\n", (int) (a), (unsigned) (b), \
             i2c_read (prua, a, b));


/** Initialize Silicon Image 9134 HDMI transmitter
 *
 * Provided for PCH media players to initialize the
 * external 9134 HDMI transmitter for audio playback.
 * To-do: Modify init code for internal 9034
 *
 * @param prua Sigma RPC handle
 * @param channels Number of enable audio output channels
 * @param frequency Output sampling rate
 * @param bits Number of rendered bits per sample
 * @param spdif True to route audio to HDMI transmitter
 * via SPDIF output, false for I2S
 * @return 0 on success, !0 on error
 */
static int sii9134_init
(
	struct RUA *prua,
	int channels,
	int frequency,
	int bits,
	int spdif
)
{
	uint8_t reg, mode = 0, lines;
	int freq_id, bits_id, fs, bs;

// --------------------------------------------------------------------

	lines = (channels + 1) >> 1;

	// Mute audio in transmitter while we apply changes

	i2c_write (prua, 0, 0x0d, i2c_read (prua, 0, 0x0d) | 0x02);

	// If TMDS isn't on, don't even bother

	if (!(i2c_read (prua, 0, 0x08) & 0x01))
	{
		g_debug ("TMDS Off - aborting audio setup");

		return -1;
	}

	// Look-up info frame & s/p-dif enumerations

	switch (bits)
	{
	default:
	case 16: bits_id = 1; bs = 2;  break;
	case 20: bits_id = 2; bs = 3;  break;
	case 24: bits_id = 3; bs = 11; break;
	}

	switch (frequency)
	{
	case 32000:  freq_id = 1; fs = 3;  break;
	case 44100:  freq_id = 2; fs = 0;  break;
	default:
	case 48000:  freq_id = 3; fs = 2;  break;
	case 88200:  freq_id = 4; fs = 8;  break;
	case 96000:  freq_id = 5; fs = 10; break;
	case 176400: freq_id = 6; fs = 12; break;
	case 192000: freq_id = 7; fs = 14; break;
	}

	// setup feed source

	if (spdif)
	{
		mode = 0x02;
	}
	else
	{
		switch (lines)
		{
		case 4:
			mode |= 0x80;
			// fall through
		case 3:
			mode |= 0x40;
			// fall through
		case 2:
			mode |= 0x20;
			// fall through
		default:
		case 1:
			mode |= 0x10;
			break;
		}
	}

	mode |= 0x01; // enable

	i2c_write (prua, 8, 0x14, mode & 0xfe);
	i2c_write (prua, 8, 0x02, 0x01); // mclk=256
	i2c_write (prua, 8, 0x15, spdif ? 0x00 : 0x02); // source
	i2c_write (prua, 8, 0x19, 0x09);
	i2c_write (prua, 8, 0x1c, (channels == 1) ? 0xe5 : 0xe4); // mono map
	i2c_write (prua, 8, 0x1d, 0x40); // hbr disable

	// set 40 bit IEC-60958 channel status

	i2c_write (prua, 8, 0x1e, 0x00);
	i2c_write (prua, 8, 0x1f, 0x19);
	i2c_write (prua, 8, 0x20, channels << 4);
	i2c_write (prua, 8, 0x21, fs);
	i2c_write (prua, 8, 0x22, ((15 - fs) << 4) | bs);

	i2c_write (prua, 8, 0x23, 0x10); // no scale factor
	i2c_write (prua, 8, 0x24, bs);

	reg = i2c_read (prua, 8, 0x2f);

	if (channels > 2)
		reg |= 0x02;
	else
		reg &= ~0x02;

	i2c_write (prua, 8, 0x2f, reg);

	// Flush fifos

	i2c_write (prua, 0, 0x05, 0x02);
	i2c_write (prua, 0, 0x05, 0x00);

	// Program in audio info frame

	i2c_write (prua, 8, 0x80, 0x84); // HDMI+audio info pkt type
	i2c_write (prua, 8, 0x81, 0x01); // packet version
	i2c_write (prua, 8, 0x82, 0x0a); // packet len
 	i2c_write (prua, 8, 0x83, 0x70);
	i2c_write (prua, 8, 0x84, (channels - 1) | (spdif ? 0 : 0x10));
	i2c_write (prua, 8, 0x85, spdif ? 0 : (freq_id << 2) | bits_id);
	i2c_write (prua, 8, 0x86, 0x00); // MSN coding type
	i2c_write (prua, 8, 0x87, (channels > 2) ? 0x10 : 0x00);
	i2c_write (prua, 8, 0x88, 0x00); // downmix and atten control
	i2c_write (prua, 8, 0x89, 0x00);

	// Enable transmission of audio info frame w/ repeat

	i2c_write (prua, 8, 0x3e, i2c_read (prua, 8, 0x3e) & 0x30);

	// g2g - enable and unmute transmitter audio

	i2c_write (prua, 8, 0x14, mode);
	i2c_write (prua, 0, 0x0d, i2c_read (prua, 0, 0x0d) & ~0x02);

	return 0;

} // end of "sii9134_init ()"


/** Test open the default device
 *
 * Test if this plugin can provide a default output, in case
 * none has been configured.  For the Sigma platform we always
 * assume true even though something - in theory - can have the
 * second logical decoder open on the first DSP.  In that case
 * the open will ultimately fail.  However the idea here is
 * never to require a valid configuration be present to function.
 *
 * @return Always true
 */
static bool smp86xx_test_default_device
(
	void
)
{
// --------------------------------------------------------------------

	return true;

} // end of "smp86xx_test_default_device ()"


/** Allocate an available STC module instance
 *
 * Locate and open an available STC module instance used
 * for pacing samples to the output I2S clocking.
 *
 * @param audio Driver context block
 * @param error Location to store error messages if they
 * occured
 * @return 0 on success, !0 on error
 */
static int _stc_init
(
	smp86xx_audio_t *audio,
	GError **error
)
{
	RMstatus sts;
	RMuint32 catagory, count = 0, i;

	struct STC_Speed_type speed;
	struct STC_Open_type profile;
	struct STC_Time_type stc_time;

// --------------------------------------------------------------------

	catagory = STC;

	RUAExchangeProperty (audio->prua, Enumerator,
	                     RMEnumeratorPropertyID_CategoryIDToNumberOfInstances,
	                     PROPERTY_ARG_REF (catagory),
	                     PROPERTY_ARG_REF (count));
	if (!count)
		return -1;

	memset (&profile, 0, sizeof (profile));

	profile.master                = Master_STC;//Audio;
	profile.stc_timer_id          = NO_TIMER;
	profile.stc_time_resolution   = STC_RESOLUTION;
	profile.video_timer_id        = NO_TIMER;
	profile.video_time_resolution = STC_RESOLUTION;
	profile.audio_time_resolution = STC_RESOLUTION;
	profile.audio_offset          = 0;
	profile.video_offset          = 0;

	// Find an available STC

	for (i = 1; i < count; i++)
	{
		profile.audio_timer_id = (i * 3) + 0;
		profile.video_timer_id = (i * 3) + 1;
		profile.stc_timer_id   = (i * 3) + 2;

		sts = RUASetProperty (audio->prua, EMHWLIB_MODULE (STC, i),
		                      RMSTCPropertyID_Open,
		                      PROPERTY_ARG_REF (profile), 0);

		if (RMSUCCEEDED (sts))
			break;
	}

	if (i == count)
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to allocate an available STC timer");
		return -1;
	}

	audio->stc_id = EMHWLIB_MODULE (STC, i);
	audio->timer_id = profile.audio_timer_id;

	g_debug ("Opened STC module instance %d\n", (int) i);

	// Make sure speed is 1:1, value 0

	speed.nominator   = 1;
	speed.denominator = 1;
	speed.stc_filter  = NoFilter;
	speed.stc_correction_method   = NoCorrection;
	speed.video_correction_method = NoCorrection;
	speed.audio_correction_method = NoCorrection;

	SETDP (audio->stc_id, RMSTCPropertyID_Speed, speed);

	stc_time.time = 0;
	stc_time.time_resolution = STC_RESOLUTION;

	SETDP (audio->stc_id, RMSTCPropertyID_Time, stc_time);

	return 0;

} // end of "_stc_init ()"


/** Configure and initialize driver
 *
 * Initializes driver context block, loads configuration
 * parameters, and precomputes needed constants.
 *
 * @param audio_format the configured audio format, or NULL if
 * none is configured
 * @param param the configuration section, or NULL if there is
 * no configuration
 * @param error location to store the error occuring, or NULL
 * to ignore errors
 * @return NULL on error, or a pointer to driver context
 */
static void *smp86xx_init
(
	G_GNUC_UNUSED const struct audio_format *audio_format,
	const struct config_param *param,
	G_GNUC_UNUSED GError **error
)
{
	smp86xx_audio_t *audio;

// --------------------------------------------------------------------

	audio = g_new (smp86xx_audio_t, 1);

	memset (audio, 0, sizeof (smp86xx_audio_t));

	audio->sp_timeout = (PENDING_RETRY_TIMEOUT * sysconf (_SC_CLK_TCK)) / 1000;

	// Fill in access parameters from config file

	audio->engine_id =
		EMHWLIB_MODULE (AudioEngine,
		                config_get_block_unsigned (param, "engine", DEFAULT_ENGINE));

	audio->decoder_id =
		EMHWLIB_MODULE (AudioDecoder,
		                config_get_block_unsigned (param, "decoder", DEFAULT_DECODER));

	audio->dram_bank =
		config_get_block_unsigned (param, "ramdac", DEFAULT_DRAM_BANK);

	audio->pcm_line_count =
		config_get_block_unsigned (param, "line_count", DEFAULT_PCM_LINE_COUNT);

	audio->fifo_duration =
		config_get_block_unsigned (param, "fifo_us", DEFAULT_BITSTREAM_FIFO_SIZE);

	audio->hdmi_output =
		config_get_block_string (param, "hdmi", NULL);

	return audio;

} // end of "smp86xx_init ()"


/** Perform synchronous wait for command completion
 *
 * Commanded state changes via emhwlib RPC call are not
 * synchronous.  Therefore we must wait on a command
 * completion event before moving forward
 *
 * @param command Command code to submit and wait for completion
 * @return Sigma status code
 */
static RMstatus _sync_audio_command
(
	smp86xx_audio_t *audio,
	enum AudioDecoder_Command_type command
)
{
#define COMMAND_COMPLETE_TIMEOUT 2000000 // in us

	RMstatus sts;
	RMuint32 result;

// --------------------------------------------------------------------

	while (1)
	{
		sts = RUASetProperty (audio->prua, audio->decoder_id,
		                      RMAudioDecoderPropertyID_Command,
		                      PROPERTY_ARG_REF (command), 0);

		if (RMSUCCEEDED (sts))
			return sts;

		if ((sts == RM_PENDING) || (sts == RM_ERROR_COMMAND_PENDING))
		{
			struct RUAEvent event;

			event.ModuleID = audio->decoder_id;
			event.Mask     = SOFT_IRQ_EVENT_COMMANDCOMPLETION;

			return RUAWaitForMultipleEvents (audio->prua, &event, 1,
			                                 COMMAND_COMPLETE_TIMEOUT, &result);
		}
	}

	return RM_OK;

} // end of "_sync_audio_command ()"


/** Shutdown and cleanup audio decoder
 *
 * Insures the Sigma audio decoder task instance is
 * closed and all associated resources are free'd
 *
 * @param audio Driver context block
 */
static void _audio_decoder_cleanup
(
	smp86xx_audio_t *audio
)
{
	RMstatus sts;

// --------------------------------------------------------------------

	if (audio->playing)
	{
		sts = _setdp (audio, audio->stc_id, RMSTCPropertyID_Stop, NULL, 0);

		g_debug ("Stopped STC - %s", RMstatusToString (sts));

		sts = _sync_audio_command (audio, AudioDecoder_Command_Stop);

		g_debug ("Stopped audio decoder - %s", RMstatusToString (sts));

		audio->playing = false;
	}

	if (audio->opened)
	{
		sts = _setdp (audio, audio->decoder_id, RMAudioDecoderPropertyID_Close, NULL, 0);

		g_debug ("Closed audio decoder - %s", RMstatusToString (sts));

		audio->opened = false;
	}

	if (audio->mem_ptr)
	{
		RUAUnMap (audio->prua, audio->mem_ptr, audio->mem_size);
		RUAUnLock (audio->prua, audio->mem_address, audio->mem_size);

		audio->mem_ptr = NULL;
		audio->fifo    = NULL;
		audio->input   = NULL;
	}

	if (audio->mem_address)
	{
		RUAFree (audio->prua, audio->mem_address);

		g_debug ("Free'd 0x%08lX bytes in uncached DRAM%d",
		         audio->mem_size, audio->dram_bank);

		audio->mem_size    = 0;
		audio->mem_address = 0;
	}

	// Make sure STC is closed

	if (audio->stc_id)
	{
		sts = _setdp (audio, audio->stc_id, RMSTCPropertyID_Close, NULL, 0);

		g_debug ("Closed STC - %s", RMstatusToString (sts));

		audio->stc_id = 0;
	}

	// Insure decoder memory is cleaned up
#if 0
	if (audio->shared_address)
	{
		SETDP (audio->engine_id, RMAudioEnginePropertyID_DecoderSharedMemory, audio->shared_saved);

		RUAUnLock (audio->prua, 
		RUAFree (audio->prua, audio->shared_address);

		audio->shared_address = 0;
	}
#endif

	// Clean up RUA handle

	if (audio->prua)
	{
		RUADestroyInstance (audio->prua);

		audio->prua = NULL;
	}

	return;

} // end of "_audio_decoder_cleanup ()"


/** Open audio decoder
 *
 * Open the audio decoder with the given sample rate, bits per
 * sample and channel count.
 *
 * @param data Driver context block
 * @param audio_format the audio format in which data is going
 * to be delivered; may be modified by the plugin
 * @param error location to store the error occuring, or NULL
 * to ignore errors
 * @return false on error, true on success
 */
static bool smp86xx_open
(
	void *data,
	struct audio_format *audio_format,
	GError **error
)
{
	smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

	RMstatus sts;
	RMuint32 address = 0, i;

	uint64_t fifo_size;

	struct AudioEngine_I2SConfig_type i2s;
	enum AudioEngine_SpdifOut_type spdif_out;
	enum AudioEngine_SerialOut_type i2s_enable;

	struct AudioDecoder_Open_type profile; 
	struct AudioDecoder_DRAMSize_in_type dram_in;
	struct AudioDecoder_DRAMSize_out_type dram_out;

	enum AudioDecoder_Codec_type codec;
	struct AudioDecoder_PcmCdaParameters_type pcm_parameters;

	struct AudioEngine_DecoderSharedMemory_type shared;

// --------------------------------------------------------------------

	// Configure audio format

	memset (&pcm_parameters, 0, sizeof (pcm_parameters));

	switch (audio_format->channels)
	{
	case 1:
		pcm_parameters.ChannelAssign  = PcmCda1_C;
		pcm_parameters.OutputChannels = Audio_Out_Ch_C;
		pcm_parameters.OutputLfe      = FALSE;
		break;

	case 2:
		pcm_parameters.ChannelAssign  = PcmCda2_LR;
		pcm_parameters.OutputChannels = Audio_Out_Ch_LR;
		pcm_parameters.OutputLfe      = FALSE;
		break;

	case 3:
		pcm_parameters.ChannelAssign  = PcmCda3_LfRfC;
		pcm_parameters.OutputChannels = Audio_Out_Ch_LCR;
		pcm_parameters.OutputLfe      = FALSE;
		break;

	case 4:
		pcm_parameters.ChannelAssign  = PcmCda4_LfRfCLfe;
		pcm_parameters.OutputChannels = Audio_Out_Ch_LCR;
		pcm_parameters.OutputLfe      = TRUE;
		break;

	case 5:
		pcm_parameters.ChannelAssign  = PcmCda5_LfRfCLsRs;
		pcm_parameters.OutputChannels = Audio_Out_Ch_LCRLsRs;
		pcm_parameters.OutputLfe      = FALSE;
		break;

	case 6:
		pcm_parameters.ChannelAssign  = PcmCda6_LfRfCLfeLsRs;
		pcm_parameters.OutputChannels = Audio_Out_Ch_LCRLsRs;
		pcm_parameters.OutputLfe      = TRUE;
		break;

	default:
		g_set_error (error, _output_quark (), 0,
		             "Unsupported channel count %d", audio_format->channels);
		return false;
	}

	switch (audio_format->bits)
	{
	case 8:
	case 16:
		break; // valid

	case 24:

		/*
		 * The Sigma platform will do 24-bit just fine, however
		 * it expects packed samples and MPD delivers the samples
		 * signed extended to 32-bit.  I understand why it is
		 * delivered this way due to dword alignment, however IMO
		 * MPD is broken in calling this a 'true 24-bit' format.
		 * So I'm falling through to forcing 16-bit output below
		 * where packing naturally equals alignment
		 */

	default:
		audio_format->bits = 16;
		break; // override
	}

	pcm_parameters.BitsPerSample     = audio_format->bits;
	pcm_parameters.SamplingFrequency = audio_format->sample_rate;
	pcm_parameters.MsbFirst          = FALSE;
	pcm_parameters.OutputDualMode    = DualMode_Stereo;
	pcm_parameters.OutputSurround20  = SurroundAsStream;
	pcm_parameters.OutputSpdif       = OutputSpdif_Uncompressed;
	pcm_parameters.SignedPCM         = TRUE;
	pcm_parameters.BassMode          = 0;

	// Open RPC handle

	sts = RUACreateInstance (&audio->prua, 0);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to open RUA handle - %s", RMstatusToString (sts));

		g_free (audio);

		return NULL;
	}

	// Zero initial time reference

	if (_stc_init (audio, error))
	{
		// logged internally

		_audio_decoder_cleanup (audio);

		return false;
	}

	// Setup audio endine

	sts = SETDP (audio->engine_id,
	             RMAudioEnginePropertyID_SampleFrequency, audio_format->sample_rate);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to set sampling frequency to %u Hz - %s",
		             audio_format->sample_rate, RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	spdif_out = AudioEngine_SpdifOut_Active;
		
	SETDP (audio->engine_id, RMAudioEnginePropertyID_SpdifOut, spdif_out);

	// Set S/P-DIF channel status

	// TBD

	// I2S Config
		
	i2s.DataAlignment   = 1;
	i2s.SClkInvert      = TRUE;
	i2s.FrameInvert     = TRUE;
	i2s.MSBFirst        = TRUE;
	i2s.SampleSize16Bit = FALSE;
	
	SETDP (audio->engine_id, RMAudioEnginePropertyID_I2SConfig, i2s);

	i2s_enable = AudioEngine_SerialOut_SO_ENABLE;

	SETDP (audio->engine_id, RMAudioEnginePropertyID_SerialOut, i2s_enable);

	// Create and lock shared memory if not already created

	memset (&shared, 0, sizeof (shared));

	RUAGetProperty (audio->prua, audio->engine_id, RMAudioEnginePropertyID_DecoderSharedMemory, &shared, sizeof(shared));

	if (!shared.Address)
	{
		struct AudioEngine_DecoderSharedMemoryInfo_in_type info_in; 
		struct AudioEngine_DecoderSharedMemoryInfo_out_type info_out;

		memset (&info_in, 0, sizeof (info_in));

		RUAExchangeProperty (audio->prua, audio->engine_id, RMAudioEnginePropertyID_DecoderSharedMemoryInfo, &info_in, sizeof(info_in), &info_out, sizeof(info_out));

		shared.Size = info_out.DecoderSharedSize;

		if (shared.Size)
		{
			shared.Address = RUAMalloc (audio->prua, audio->dram_bank, RUA_DRAM_UNCACHED, shared.Size);

			if (shared.Address)
			{
				audio->shared_address = shared.Address;
				audio->shared_size = shared.Size;

				SETDP (audio->engine_id, RMAudioEnginePropertyID_DecoderSharedMemory, shared);

				RUALock (audio->prua, audio->shared_address, audio->shared_size);
			}
		}
	}

	// Size up decoder ram requirements & allocate

	fifo_size  = audio_format->sample_rate;
	fifo_size *= (audio_format->bits / 8);
	fifo_size *= audio_format->channels;
	fifo_size *= (uint64_t) audio->fifo_duration; // in us
	fifo_size /= (uint64_t) 1000000;

	// round up to next log2

	for (i = 31; i; i--)
	{
		if (fifo_size & (1 << i))
		{
			fifo_size = 1 << (i + 1);

			break;
		}
	}

	audio->fifo_size = fifo_size;

	g_debug ("FIFO Size = %d bytes", (int) fifo_size);

	memset (&dram_in, 0, sizeof (dram_in));
	
	dram_in.MaxChannelOutCount = 12;//audio_format->channels;
	dram_in.PCMLineCount       = audio->pcm_line_count;
	dram_in.BitstreamFIFOSize  = fifo_size;
	dram_in.XferFIFOCount      = 0;

	RUAExchangeProperty (audio->prua, audio->decoder_id,
	                     RMAudioDecoderPropertyID_DRAMSize,
	                     PROPERTY_ARG_REF (dram_in),
	                     PROPERTY_ARG_REF (dram_out));

	memset (&profile, 0, sizeof (profile));

	profile.MaxChannelOutCount = dram_in.MaxChannelOutCount;
	profile.PCMLineCount       = dram_in.PCMLineCount;
	profile.BitstreamFIFOSize  = dram_in.BitstreamFIFOSize;
	profile.XferFIFOCount      = dram_in.XferFIFOCount;
	profile.CachedSize         = 0;
	profile.CachedAddress      = 0;
	profile.UncachedSize       = dram_out.UncachedSize;
	profile.DemuxProgramId     = 0;
	profile.TimerId            = audio->timer_id;

	profile.UncachedAddress =
		RUAMalloc (audio->prua, audio->dram_bank, RUA_DRAM_UNCACHED, dram_out.UncachedSize);

	if (!profile.UncachedAddress)
	{
		g_set_error (error, _output_quark (), 0,
		             "Could not allocate 0x%08lX bytes in uncached DRAM%d",
		             dram_out.UncachedSize, audio->dram_bank);

		_audio_decoder_cleanup (audio);

		return false;
	}

	g_debug ("Allocated 0x%08lX bytes in uncached DRAM%d",
	         dram_out.UncachedSize, audio->dram_bank);

	audio->mem_size    = dram_out.UncachedSize;
	audio->mem_address = profile.UncachedAddress;

	// Map the input sample fifo into virtual process space

	sts = RUALock (audio->prua, audio->mem_address, audio->mem_size);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to lock-in audio fifo - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	audio->mem_ptr = (uint8_t *) RUAMap (audio->prua, audio->mem_address, audio->mem_size);

	if (!audio->mem_ptr)
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to map audio fifo into virtual memory");

		RUAUnLock (audio->prua, audio->mem_address, audio->mem_size);

		_audio_decoder_cleanup (audio);

		return false;
	}

	// Open decoder and set codec/parameters

	sts = SETDP (audio->decoder_id, RMAudioDecoderPropertyID_Open, profile);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to open audio decoder - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	audio->opened = true;

	sts = _sync_audio_command (audio, AudioDecoder_Command_Uninit);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to uninit audio decoder - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	g_debug ("Audio decoder reset");

	codec = AudioDecoder_Codec_PCM;
	 
	sts = SETDP (audio->decoder_id, RMAudioDecoderPropertyID_Codec, codec);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to set audio codec to PCM - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	sts = SETDP (audio->decoder_id,
	             RMAudioDecoderPropertyID_PcmCdaParameters, pcm_parameters);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to set PCM properties - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	sts = _sync_audio_command (audio, AudioDecoder_Command_Init);

	if (RMFAILED (sts))
	{
		g_set_error (error, _output_quark (), 0,
		             "Unable to initialize audio decoder - %s", RMstatusToString (sts));

		_audio_decoder_cleanup (audio);

		return false;
	}

	g_debug ("Audio decoder initialized for PCM playback");

	// Setup FIFO pointers

	RUAGetProperty (audio->prua, audio->decoder_id,
	                RMAudioDecoderPropertyID_BtsFIFO,
	                PROPERTY_ARG_REF (address));

	audio->fifo = (volatile struct gbus_fifo *)
		(audio->mem_ptr + (address - audio->mem_address));

	audio->input = (int8_t *)
		(audio->mem_ptr + (audio->fifo->base - audio->mem_address));

	// Setup HDMI

	if (audio->hdmi_output && !strcasecmp (audio->hdmi_output, "sii9134"))
		sii9134_init (audio->prua, audio_format->channels,
		              audio_format->sample_rate, audio_format->bits, TRUE);

	return true;

} // end of "smp86xx_open ()"


/** Play a chunk of audio data.
 *
 * Write a block of samples to the audio decoder input
 * fifo.  The data chunk is copied into the decoder input
 * fifo atomically from the decoder's point of view (ie.
 * the write pointer is advanced one time at end).
 *
 * @param data Local driver context
 * @param chunk Incoming sample buffer
 * @param len Size of sample data in bytes
 * @param error location to store the error occuring, or NULL
 * to ignore errors
 * @return the number of bytes played, or 0 on error
 */
static size_t smp86xx_play
(
	void *data,
	const void *chunk,
	size_t len,
	G_GNUC_UNUSED GError **error
)
{
	smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

	uint32_t size, wr, rd, bytes, count = 0, reset_flag = 0;

	const uint8_t *input = (const uint8_t *) chunk;

// --------------------------------------------------------------------

	wr = audio->fifo->wr;

	while (len)
	{
		rd = audio->fifo->rd;

		if (wr >= rd)
			size = audio->fifo_size - wr - (rd ? 0 : 1);
		else
			size = rd - wr - 1;

		bytes = (size > len) ? len : size;

		if (!bytes)
		{
			usleep (50000);

// need a long term timeout here



			continue;
		}

		reset_flag = 0;

		memcpy (&audio->input[wr], input, bytes);

		wr     = (wr + bytes) % audio->fifo_size;
		input += bytes;
		len   -= bytes;
		count += bytes;
	}

	audio->fifo->wr = wr;

	if (!audio->playing || !audio->fifo->rd)
	{
		RMstatus sts;

		sts = _sync_audio_command (audio, AudioDecoder_Command_Play);

		g_debug ("Playing audio decoder - %s", RMstatusToString (sts));

		sts = _setdp (audio, audio->stc_id, RMSTCPropertyID_Play, NULL, 0);

		g_debug ("Playing STC - %s", RMstatusToString (sts));

		audio->playing = true;
	}


#if 0
	g_debug ("%u %u %u",
			  (unsigned int) audio->fifo->size,
			  (unsigned int) audio->fifo->rd,
			  (unsigned int) audio->fifo->wr);
#endif
	return count;

} // end of "smp86xx_play ()"


/** Display meta data block
 *
 * Display in-band meta-tags in the audio stream
 *
 * @param data Local driver context
 * @param tag in-band meta data content
 */
static void smp86xx_send_tag
(
	G_GNUC_UNUSED void *data,
	const struct tag *tag
)
{
	unsigned int i;

	const char *label;

	//smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

// --------------------------------------------------------------------

	for (i = 0; i < tag->num_items; i++)
	{
		label = NULL;

		switch (tag->items[i]->type)
		{
		case TAG_ITEM_ARTIST:       label = "Artist"; break;
		case TAG_ITEM_ALBUM:        label = "Album"; break;
		case TAG_ITEM_ALBUM_ARTIST: label = "Album Artist"; break;
		case TAG_ITEM_TITLE:        label = "Title"; break;
		case TAG_ITEM_TRACK:        label = "Track"; break;
		case TAG_ITEM_NAME:         label = "Name"; break;
		case TAG_ITEM_GENRE:        label = "Genre"; break;
		case TAG_ITEM_DATE:         label = "Date"; break;
		case TAG_ITEM_COMPOSER:     label = "Composer"; break;
		case TAG_ITEM_PERFORMER:    label = "Performer"; break;
		case TAG_ITEM_COMMENT:      label = "Comment"; break;
		case TAG_ITEM_DISC:         label = "Disc"; break;

		default:
			g_debug ("tag(+%ds): Unhandled %d", tag->time, tag->items[i]->type);
			break;
		}

		if (label)
		{
			//g_debug ("tag(+%ds): %s - %s", tag->time, label, (char *) &tag->items[i]->value[0]);
			g_debug ("tag(+%ds): %s", tag->time, label);
		}
	}

	return;

} // end of "smp86xx_send_tag ()"


/** Clear output sample buffer
 *
 * Try to cancel data which may still be in the device's
 * buffers.
 *
 * @param data Driver context block
 */
static void smp86xx_cancel
(
	void *data
)
{
	smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

// --------------------------------------------------------------------

	audio->fifo->wr = audio->fifo->rd = 0;

	return;

} // end of "smp86xx_cancel ()"


/** Shutdown and cleanup audio decoder
 *
 * Insures the Sigma audio decoder task instance is
 * closed and all associated resources are free'd
 *
 * @param data Driver context block
 */
static void smp86xx_close
(
	void *data
)
{
	smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

// --------------------------------------------------------------------

	_audio_decoder_cleanup (audio);

	return;

} // end of "smp86xx_close ()"


/** Terminate access to smp86xx driver
 *
 * Make sure audio device is closed and free all resources
 * associated with driver
 */
static void smp86xx_term
(
	void *data
)
{
	smp86xx_audio_t *audio = (smp86xx_audio_t *) data;

// --------------------------------------------------------------------

	_audio_decoder_cleanup (audio);

	g_free (audio);

	return;

} // end of "smp86xx_term ()"


extern const struct mixer_plugin smp86xx_mixer;

const struct audio_output_plugin smp86xx =
{
	.name      = "smp86xx",
	.test_default_device = smp86xx_test_default_device,
	.init      = smp86xx_init,
	.open      = smp86xx_open,
	.send_tag  = smp86xx_send_tag,
	.play      = smp86xx_play,
	.cancel    = smp86xx_cancel,
	.close     = smp86xx_close,
	.finish    = smp86xx_term,
	.mixer_plugin = &smp86xx_mixer,
};


// Sigma portablility wrapper functions

#include <stdio.h>
#include <stdlib.h>

int verbose_stderr = 1;
RMbool manutest = FALSE;

void *RMMalloc (RMuint32 size)
{
	return malloc (size);
}

void RMFree (void *ptr)
{
	free (ptr);
}

void *RMMemset (void *s, RMuint8 c, RMuint32 n)
{
	return memset (s, c, n);
}

void *RMMemcpy (void *dest, const void *src, RMuint32 n)
{
	return memcpy (dest, src, n);
}

RMint32 RMMemcmp (const void *s1, const void *s2, RMuint32 n)
{
	return memcmp (s1, s2, n);
}

void *RMCalloc (RMuint32 nmemb, RMuint32 size)
{
	return calloc (nmemb, size);
}

void RMDBGLOG_implementation
(
	RMbool active,
	G_GNUC_UNUSED const RMascii *filename,
	G_GNUC_UNUSED RMint32 line,
	const RMascii *text,
	...
)
{
	va_list ap;

// --------------------------------------------------------------------

	if (!active)
		return;

	va_start(ap, text);
	vprintf(text, ap);
	va_end(ap);

	fflush(stdout);

	return;
}

void RMDBGPRINT_implementation
(
	RMbool active,
	G_GNUC_UNUSED const RMascii *filename,
	G_GNUC_UNUSED RMint32 line,
	const RMascii *text,
	...
)
{
	va_list ap;

// --------------------------------------------------------------------

	if (!active)
		return;

	va_start(ap, text);
	vprintf(text, ap);
	va_end(ap);

	fflush(stdout);

	return;
}
